import unittest
import uuid

import gtirb


class NodeFromUUIDTest(unittest.TestCase):
    def test_get_by_uuid_failed(self):
        bad_id = uuid.uuid4()
        node = gtirb.IR().get_by_uuid(bad_id)
        self.assertIsNone(node)

    def test_get_by_uuid(self):
        ir1 = gtirb.IR()
        ir2 = gtirb.IR()

        # test nodes from one IR don't pollute the cache of other IRs
        self.assertIsNone(ir1.get_by_uuid(ir2.uuid))
        self.assertIsNone(ir2.get_by_uuid(ir1.uuid))
        self.assertEqual(ir1.get_by_uuid(ir1.uuid), ir1)
        self.assertEqual(ir2.get_by_uuid(ir2.uuid), ir2)

        m = gtirb.Module(name="M")
        m.ir = ir1
        s = gtirb.Section()
        s.module = m
        bi = gtirb.ByteInterval()
        bi.section = s
        b = gtirb.CodeBlock()
        b.byte_interval = bi
        sym = gtirb.Symbol("test")
        sym.module = m

        # test all nodes are lookupable by any other node type
        self.assertEqual(ir1.get_by_uuid(m.uuid), m)
        self.assertEqual(ir1.get_by_uuid(s.uuid), s)
        self.assertEqual(ir1.get_by_uuid(bi.uuid), bi)
        self.assertEqual(ir1.get_by_uuid(b.uuid), b)
        self.assertEqual(ir1.get_by_uuid(sym.uuid), sym)

        self.assertEqual(m.ir.get_by_uuid(m.uuid), m)
        self.assertEqual(m.ir.get_by_uuid(s.uuid), s)
        self.assertEqual(m.ir.get_by_uuid(bi.uuid), bi)
        self.assertEqual(m.ir.get_by_uuid(b.uuid), b)
        self.assertEqual(m.ir.get_by_uuid(sym.uuid), sym)

        self.assertEqual(s.ir.get_by_uuid(m.uuid), m)
        self.assertEqual(s.ir.get_by_uuid(s.uuid), s)
        self.assertEqual(s.ir.get_by_uuid(bi.uuid), bi)
        self.assertEqual(s.ir.get_by_uuid(b.uuid), b)
        self.assertEqual(s.ir.get_by_uuid(sym.uuid), sym)

        self.assertEqual(bi.ir.get_by_uuid(m.uuid), m)
        self.assertEqual(bi.ir.get_by_uuid(s.uuid), s)
        self.assertEqual(bi.ir.get_by_uuid(bi.uuid), bi)
        self.assertEqual(bi.ir.get_by_uuid(b.uuid), b)
        self.assertEqual(bi.ir.get_by_uuid(sym.uuid), sym)

        self.assertEqual(b.ir.get_by_uuid(m.uuid), m)
        self.assertEqual(b.ir.get_by_uuid(s.uuid), s)
        self.assertEqual(b.ir.get_by_uuid(bi.uuid), bi)
        self.assertEqual(b.ir.get_by_uuid(b.uuid), b)
        self.assertEqual(b.ir.get_by_uuid(sym.uuid), sym)

        self.assertEqual(sym.ir.get_by_uuid(m.uuid), m)
        self.assertEqual(sym.ir.get_by_uuid(s.uuid), s)
        self.assertEqual(sym.ir.get_by_uuid(bi.uuid), bi)
        self.assertEqual(sym.ir.get_by_uuid(b.uuid), b)
        self.assertEqual(sym.ir.get_by_uuid(sym.uuid), sym)

        # test removing a node removes all children as well
        bi.section = None

        self.assertEqual(ir1.get_by_uuid(m.uuid), m)
        self.assertEqual(ir1.get_by_uuid(s.uuid), s)
        self.assertIsNone(ir1.get_by_uuid(bi.uuid))
        self.assertIsNone(ir1.get_by_uuid(b.uuid))
        self.assertEqual(ir1.get_by_uuid(sym.uuid), sym)

        bi.section = s

        self.assertEqual(ir1.get_by_uuid(m.uuid), m)
        self.assertEqual(ir1.get_by_uuid(s.uuid), s)
        self.assertEqual(ir1.get_by_uuid(bi.uuid), bi)
        self.assertEqual(ir1.get_by_uuid(b.uuid), b)
        self.assertEqual(ir1.get_by_uuid(sym.uuid), sym)

    def test_remove_from_cache(self):
        ir = gtirb.IR()
        m = gtirb.Module(name="M")
        m.ir = ir
        s = gtirb.Section()

        s.module = m
        self.assertEqual(ir.get_by_uuid(s.uuid), s)

        s.module = None
        self.assertIsNone(ir.get_by_uuid(s.uuid))

        m.sections.discard(s)
        self.assertIsNone(ir.get_by_uuid(s.uuid))


if __name__ == "__main__":
    unittest.main()



import io
import unittest

import gtirb
from helpers import create_interval_etc


class SymbolicExpressionAttributes(unittest.TestCase):
    def test_unknown_attributes(self):
        ir, m, s, bi = create_interval_etc(address=0x1000, size=5)
        sym = gtirb.Symbol(name="foo")
        m.symbols.add(sym)
        expr = gtirb.SymAddrConst(
            0, sym, {gtirb.SymbolicExpression.Attribute.GOT, 0xBEEF}
        )
        bi.symbolic_expressions[0] = expr

        out = io.BytesIO()
        ir.save_protobuf_file(out)
        out.seek(0)

        ir = gtirb.IR.load_protobuf_file(out)
        bi = next(ir.byte_intervals)
        expr = bi.symbolic_expressions[0]

        self.assertTrue(
            gtirb.SymbolicExpression.Attribute.GOT in expr.attributes
        )
        self.assertTrue(0xBEEF in expr.attributes)



import unittest

import gtirb
from helpers import SearchScope, create_interval_etc, parameterize_one


class BlocksOnTests(unittest.TestCase):
    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_simple(self, scope):
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=3, byte_interval=bi)
        code_block2 = gtirb.CodeBlock(offset=3, size=1, byte_interval=bi)

        found = set(search_in.byte_blocks_on(0x1001))
        self.assertEqual(found, {code_block})

        found = set(search_in.byte_blocks_on(0x1003))
        self.assertEqual(found, {code_block2})

        # Change the offset to verify we update the index
        code_block.offset = 2
        found = set(search_in.byte_blocks_on(0x1000))
        self.assertEqual(found, set())

        found = set(search_in.byte_blocks_on(0x1002))
        self.assertEqual(found, {code_block})

        # Discard the block to verify we update the index
        bi.blocks.discard(code_block)
        found = set(search_in.byte_blocks_on(0x1002))
        self.assertEqual(found, set())

        # Now add it back to verify we update the index
        bi.blocks.add(code_block)
        found = set(search_in.byte_blocks_on(0x1002))
        self.assertEqual(found, {code_block})

    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_zero(self, scope):
        ir, m, s, bi = create_interval_etc(address=0, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=3, byte_interval=bi)

        found = set(search_in.byte_blocks_on(0))
        self.assertEqual(found, {code_block})

    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_with_overlapping_blocks(self, scope):
        "Test that byte_blocks_on returns all blocks that overlap an address"
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi)
        code_block2 = gtirb.CodeBlock(offset=0, size=2, byte_interval=bi)
        code_block3 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi)

        found = set(search_in.byte_blocks_on(0x1001))
        self.assertEqual(found, {code_block2, code_block3})

    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_with_zero_sized_blocks(self, scope):
        "Test that byte_blocks_on doesn't find zero-sized blocks"
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=3, byte_interval=bi)
        code_block2 = gtirb.CodeBlock(offset=1, size=0, byte_interval=bi)
        code_block3 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi)

        found = set(search_in.byte_blocks_on(0x1001))
        self.assertEqual(found, {code_block, code_block3})

        found = set(search_in.byte_blocks_on(range(0x1001, 0x1002)))
        self.assertEqual(found, {code_block, code_block3})

    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_with_range(self, scope):
        "Test that byte_blocks_on handles ranges"
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=2, byte_interval=bi)
        code_block2 = gtirb.CodeBlock(offset=2, size=1, byte_interval=bi)
        code_block3 = gtirb.CodeBlock(offset=3, size=1, byte_interval=bi)

        found = set(search_in.byte_blocks_on(range(0x1000, 0x1004)))
        self.assertEqual(found, {code_block, code_block2, code_block3})

        found = set(search_in.byte_blocks_on(range(0x1002, 0x1002)))
        self.assertEqual(found, set())

        # Passing a different step doesn't make a ton of sense, but it should
        # work.
        found = set(search_in.byte_blocks_on(range(0x1000, 0x1004, 2)))
        self.assertEqual(found, {code_block, code_block2, code_block3})

    @parameterize_one("scope", list(SearchScope))
    def test_blocks_on_with_no_address(self, scope):
        "Test that byte_blocks_on does nothing if we don't have an address"
        ir, m, s, bi = create_interval_etc(address=None, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=2, byte_interval=bi)
        code_block2 = gtirb.CodeBlock(offset=2, size=1, byte_interval=bi)
        code_block3 = gtirb.CodeBlock(offset=3, size=1, byte_interval=bi)

        found = set(search_in.byte_blocks_on(range(0x1000, 0x1004)))
        self.assertEqual(found, set())

    @parameterize_one("scope", list(SearchScope))
    def test_code_blocks_on(self, scope):
        "Test that code_blocks_on only gives back CodeBlocks"
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi)
        data_block = gtirb.DataBlock(offset=1, size=1, byte_interval=bi)

        found = set(search_in.code_blocks_on(range(0x1000, 0x1004)))
        self.assertEqual(found, {code_block})

    @parameterize_one("scope", list(SearchScope))
    def test_data_blocks_on(self, scope):
        "Test that data_blocks_on only gives back DataBlocks"
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        search_in = scope.select(ir, m, s, bi)

        code_block = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi)
        data_block = gtirb.DataBlock(offset=1, size=1, byte_interval=bi)

        found = set(search_in.data_blocks_on(range(0x1000, 0x1004)))
        self.assertEqual(found, {data_block})


class SectionBlocksOnTests(unittest.TestCase):
    def test_blocks_on_simple(self):
        s = gtirb.Section()

        bi1 = gtirb.ByteInterval(address=0x1000, size=4, section=s)
        bi1_block1 = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi1)
        bi1_block2 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi1)

        bi2 = gtirb.ByteInterval(address=0x1004, size=4, section=s)
        bi2_block1 = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi2)
        bi2_block2 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi2)

        found = set(s.byte_blocks_on(0x1001))
        self.assertEqual(found, {bi1_block2})

    def test_blocks_on_overlapping(self):
        "Test that we find the correct blocks if two byte intervals overlap"
        s = gtirb.Section()

        bi1 = gtirb.ByteInterval(address=0x1000, size=4, section=s)
        bi1_block1 = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi1)
        bi1_block2 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi1)

        bi2 = gtirb.ByteInterval(address=0x1000, size=4, section=s)
        bi2_block1 = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi2)
        bi2_block2 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi2)

        found = set(s.byte_blocks_on(0x1001))
        self.assertEqual(found, {bi1_block2, bi2_block2})

    def test_blocks_on_with_blocks_outside_bi(self):
        "Tests that we can handle byte intervals with blocks outside"
        s = gtirb.Section()

        bi1 = gtirb.ByteInterval(address=0x1000, size=1, section=s)
        bi1_block1 = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi1)
        bi1_block2 = gtirb.CodeBlock(offset=1, size=1, byte_interval=bi1)

        found = set(s.byte_blocks_on(0x1000))
        self.assertEqual(found, {bi1_block1})

        # These blocks are outside of the byte interval's declared size, so
        # either interpretation is fair game.
        found = set(s.byte_blocks_on(0x1001))
        self.assertTrue(found == set() or found == {bi1_block2})



from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Optional
from uuid import UUID

from google.protobuf.internal.containers import MessageMap

from .node import Node
from .proto import AuxData_pb2
from .serialization import Serialization
from .util import DictLike

if TYPE_CHECKING:  # pragma: no cover
    # Ignore flake8 "imported but unused" errors.
    from .ir import IR  # noqa: F401


class _LazyDataContainer:
    """
    Container that holds the raw byte stream until it is read, then releases
    it. If it is never read, then serialization skips re-encoding (and
    deserializing) the data.
    """

    def __init__(
        self,
        raw_data: bytes,
        type_name: str,
        get_by_uuid: Callable[[UUID], Optional[Node]],
    ):
        self.raw_data: Optional[bytes] = raw_data
        self.type_name = type_name
        self.get_by_uuid = get_by_uuid

    def get_data(self) -> object:
        """
        Get any pending still-serialized data, or return the passed data
        instead (the default).
        """
        assert self.raw_data is not None
        rv = AuxData.serializer.decode(
            self.raw_data, self.type_name, self.get_by_uuid
        )
        self.raw_data = None
        return rv

    def get_raw_data(self) -> bytes:
        """ """
        assert self.raw_data is not None
        return self.raw_data


class AuxData:
    """AuxData objects can be attached to the :class:`gtirb.IR` or individual
    :class:`gtirb.Module` s to store additional client-specific data in a
    portable way.

    AuxData represents a portable, language-independent manner of encoding
    rich data. To do this, all data is stored on disk as a series of bytes
    with a string describing the format of the data, called a *type name*. See
    :mod:`gtirb.serialization` for the list of all default types. Types may
    also be parameterized; for example, ``mapping<string,UUID>`` is a ``dict``
    from ``str`` objects to ``UUID`` objects. All ``AuxData`` requires
    a valid type name in order to be serialized.

    :ivar ~.data: The value stored in this AuxData.
    :ivar ~.type_name: A string describing the type of ``data``.
        Used to determine the proper codec for serializing this AuxData.
    """

    serializer: ClassVar[Serialization] = Serialization()
    """This is a :class:`gtirb.Serialization` instance, used to
    encode and decode ``data`` fields of all ``AuxData``. See
    :mod:`gtirb.serialization` for details.
    """

    def __init__(
        self,
        data: object,
        type_name: str,
        lazy_container: Optional[_LazyDataContainer] = None,
    ):
        """
        :param data: The value stored in this AuxData.
        :param type_name: A string describing the type of ``data``.
            Used to determine the proper codec for serializing this AuxData.
        :param lazy_container: An object that will lazily deserialize the
            auxdata table backing this object, or None.
        """
        self._lazy_container = lazy_container
        # _data has type Any to avoid disrupting clients want to type check
        # their use of gtirb. If _data had type object, they would have to
        # verify the element types of potentially large containers, or else
        # just subvert the type system by casting anyway.
        self._data: Any = data  # type: ignore[misc]
        self.type_name = type_name

    @property
    def data(self) -> Any:  # type: ignore[misc]
        if self._lazy_container is not None:
            self._data = self._lazy_container.get_data()
            self._lazy_container = None
        return self._data

    @data.setter
    def data(self, value: object) -> None:
        self._data = value
        self._lazy_container = None

    @classmethod
    def _from_protobuf(
        cls,
        aux_data: AuxData_pb2.AuxData,
        ir: Optional["IR"],
    ) -> "AuxData":
        """Deserialize AuxData from Protobuf. Lazy, will not perform
        deserialization until .data is accessed.

        :param aux_data: The Protobuf AuxData object.
        """

        # Defer deserialization until someone accesses .data
        assert ir
        lazy_container = _LazyDataContainer(
            aux_data.data, aux_data.type_name, ir.get_by_uuid
        )
        return cls(
            data=None,
            type_name=aux_data.type_name,
            lazy_container=lazy_container,
        )

    def _to_protobuf(self) -> AuxData_pb2.AuxData:
        """Get a Protobuf representation of the AuxData."""

        proto_auxdata = AuxData_pb2.AuxData()
        proto_auxdata.type_name = self.type_name
        # If we are serializing the same data, and the way that data is encoded
        # has not changed, then just use the already serialized copy.
        if self._lazy_container is not None and (
            self.type_name == self._lazy_container.type_name
        ):
            proto_auxdata.data = self._lazy_container.get_raw_data()
        else:
            data_stream = BytesIO()
            AuxData.serializer.encode(data_stream, self.data, self.type_name)
            proto_auxdata.data = data_stream.getvalue()
        return proto_auxdata

    def __repr__(self) -> str:
        return (
            "AuxData("
            "type_name={type_name!r}, "
            "data={data!r}, "
            ")".format(type_name=self.type_name, data=self.data)
        )


class AuxDataContainer(Node):
    """The base class for anything that holds AuxData tables; that is,
    :class:`gtirb.IR` and :class:`gtirb.Module`.

    :ivar ~.aux_data: The auxiliary data associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`.
    """

    def __init__(
        self,
        aux_data: DictLike[str, AuxData] = {},
        uuid: Optional[UUID] = None,
    ):
        """
        :param aux_data: The initial auxiliary data to be associated
            with the object, as a mapping from names to
            :class:`gtirb.AuxData`. Defaults to an empty :class:`dict`.
        :param uuid: the UUID of this ``AuxDataContainer``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        """
        super().__init__(uuid)
        self.aux_data: Dict[str, AuxData] = dict(aux_data)

    @classmethod
    def _read_protobuf_aux_data(
        cls,
        proto_container: "MessageMap[str, AuxData_pb2.AuxData]",
        ir: Optional["IR"],
    ) -> Dict[str, AuxData]:
        """
        Instead of the overrided _decode_protobuf, this method requires the
        Protobuf message to read from. AuxDataContainers need to call this
        method in their own _decode_protobuf overrides.

        :param proto_container: A Protobuf message with a field called
            ``aux_data``.
        """
        return {
            key: AuxData._from_protobuf(val, ir)
            for key, val in proto_container.items()
        }

    def _write_protobuf_aux_data(
        self, proto_container: "MessageMap[str, AuxData_pb2.AuxData]"
    ) -> None:
        """
        Instead of the overrided _to_protobuf, this method requires the
        Protobuf message to write into. AuxDataContainers need to call this
        method in their own _to_protobuf overrides.

        :param proto_container: A Protobuf message with a field called
            ``aux_data``.
        """
        for k, v in self.aux_data.items():
            proto_container[k].CopyFrom(v._to_protobuf())

    def deep_eq(self, other: object) -> bool:
        """This overrides :func:`gtirb.Node.deep_eq` to check for
        AuxData equality.

        Because the values stored by AuxData are not necessarily
        amenable to deep checking, the auxiliary data dictionaries
        stored for ``self`` and ``other`` are not deeply checked. Instead,
        they are considered to be equal if their sets of keys are equal.
        """

        if not isinstance(other, AuxDataContainer):
            return False
        if (
            self.uuid != other.uuid
            or self.aux_data.keys() != other.aux_data.keys()
        ):
            return False
        return True



"""General utilities usable by any other GTIRB submoudle."""

import itertools
import typing

import intervaltree
import typing_extensions

K = typing.TypeVar("K")
V = typing.TypeVar("V")
T = typing.TypeVar("T")
T_cov = typing.TypeVar("T_cov", covariant=True)
T_contra = typing.TypeVar("T_contra", contravariant=True)
S = typing.TypeVar("S")


DictLike = typing.Union[
    typing.Mapping[K, V],
    typing.Iterable[typing.Tuple[K, V]],
]
"""Any value that can be passed to the constructor of ``dict``;
that is, a mapping or iterable yielding key-value tuples.
"""


class GtirbError(Exception):
    pass


class DeserializationError(GtirbError):
    pass


class _SymbolicExpressionContainer(typing_extensions.Protocol[T_cov]):
    """A container of symbolic expressions at addresses."""

    def symbolic_expressions_at(
        self, addrs: typing.Union[int, range]
    ) -> typing.Iterable[T_cov]:
        ...  # pragma: no cover


class ListWrapper(typing.MutableSequence[T]):
    def __init__(self, *args: typing.Iterable[T]):
        self._data: typing.List[T] = []
        for values in args:
            for value in values:
                self.append(value)

    def _add(self, value: T) -> None:
        pass  # pragma: no cover

    def _remove(self, value: T) -> None:
        pass  # pragma: no cover

    # begin functions for ABC
    @typing.overload
    def __getitem__(self, i: int) -> T:
        ...  # pragma: no cover

    @typing.overload
    def __getitem__(self, i: slice) -> typing.MutableSequence[T]:
        ...  # pragma: no cover

    def __getitem__(
        self, i: typing.Union[int, slice]
    ) -> typing.Union[T, typing.MutableSequence[T]]:
        return self._data[i]

    @typing.overload
    def __setitem__(self, i: typing_extensions.SupportsIndex, v: T) -> None:
        ...  # pragma: no cover

    @typing.overload
    def __setitem__(self, i: slice, v: typing.Iterable[T]) -> None:
        ...  # pragma: no cover

    def __setitem__(
        self,
        i: typing.Union[typing_extensions.SupportsIndex, slice],
        v: typing.Union[T, typing.Iterable[T]],
    ) -> None:
        if isinstance(i, slice):
            assert isinstance(v, typing.Iterable)
            indices = range(*i.indices(len(self)))
            values = list(v)
        elif -len(self._data) <= i.__index__() < len(self._data):
            indices = range(i.__index__(), i.__index__() + 1)
            values = [typing.cast(T, v)]
        else:
            raise IndexError("list assignment index out of range")
        for index in indices:
            self._remove(self._data[index])
        for value in values:
            self._add(value)
        if isinstance(i, slice):
            self._data[i] = values
        else:
            self._data[i] = values[0]

    @typing.overload
    def __delitem__(self, i: int) -> None:
        ...  # pragma: no cover

    @typing.overload
    def __delitem__(self, i: slice) -> None:
        ...  # pragma: no cover

    def __delitem__(self, i: typing.Union[int, slice]) -> None:
        if isinstance(i, slice):
            indices = range(*i.indices(len(self)))
        else:
            indices = range(i, i + 1)
        for index in indices:
            self._remove(self._data[index])

        del self._data[i]

    def __len__(self) -> int:
        return len(self._data)

    def insert(self, i: int, v: T) -> None:
        self._add(v)
        return self._data.insert(i, v)

    # The version of typing.py which comes with python 3.5.2 doesn't provide
    # definitions for append or remove on MutableList, so we have to do it
    # ourselves.
    def append(self, v: T) -> None:
        self.insert(len(self), v)

    def remove(self, v: T) -> None:
        del self[self._data.index(v)]

    # extend is not in every version of Python 3, so list wrapper adds it here
    # itself.
    def extend(self, other: typing.Iterable[T]) -> None:
        for v in other:
            self.append(v)

    # end functions for ABC
    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


# A type variable for the SetWrapper's "self" type. Since type variables cannot
# have a generic bound (see https://github.com/python/mypy/issues/2756), we
# need to use Any instead. Which we then have to tell mypy to ignore.
#
# Used in __ior__.
_SetWrapperSelf = typing.TypeVar(  # type: ignore[misc]
    "_SetWrapperSelf", bound="SetWrapper[typing.Any]"
)


class SetWrapper(typing.MutableSet[T]):
    def __init__(self, *args: typing.Iterable[T]):
        self._data: typing.Set[T] = set()
        for arg in args:
            for v in arg:
                self.add(v)

    # begin functions for ABC
    def __contains__(self, v: object) -> bool:
        return v in self._data

    def __iter__(self) -> typing.Iterator[T]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def add(self, v: T) -> None:
        return self._data.add(v)

    def discard(self, v: T) -> None:
        return self._data.discard(v)

    # end functions for ABC

    # The version of typing.py which comes with python 3.5.2 doesn't provide
    # definitions for __or__ or clear on MutableSet, so we have to do it
    # ourselves.
    def __or__(
        self, other: typing.AbstractSet[S]
    ) -> typing.Set[typing.Union[T, S]]:
        return self._data | other

    # The type declaration for __ior__ in typeshed's MutableSet is problematic.
    # MutableSet requires __ior__ to accept an AbstractSet[S], and return a
    # Mutable[T|S]. This results in a type error when the result is assigned
    # back to the original variable unless S is T. Since this can't actually be
    # used if S is not T and is much easier to implement if S is T anyway, we
    # declare it to only accept AbstractSet[T]. This causes mypy to report an
    # error for incompatible override types and different return types for
    # __ior__ and __or__, which we ignore.
    def __ior__(  # type: ignore
        self: _SetWrapperSelf, other: typing.AbstractSet[T]
    ) -> _SetWrapperSelf:
        for value in other:
            self.add(value)
        return self

    def pop(self) -> T:
        it = iter(self)
        # pop is documented as raising a KeyError if it's empty, not
        # StopIteration
        try:
            result = next(it)
        except StopIteration:
            raise KeyError
        self.discard(result)
        return result

    def clear(self) -> None:
        while self:
            self.pop()

    # For whatever reason, update isn't included as part of abc.MutableSet.
    def update(self, *others: typing.Iterable[T]) -> None:
        for other in others:
            for v in other:
                self.add(v)

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


class DictWrapper(typing.MutableMapping[K, V]):
    def __init__(self, *args: DictLike[K, V]):
        self._data: typing.MutableMapping[K, V] = {}
        # Create a temporary dictionary so we can uniformly access the items
        # and add them to _data.
        temp: typing.Dict[K, V] = dict(*args)
        for i, v in temp.items():
            self[i] = v

    # begin functions for ABC
    def __getitem__(self, i: K) -> V:
        return self._data[i]

    def __setitem__(self, i: K, v: V) -> None:
        self._data[i] = v

    def __delitem__(self, i: K) -> None:
        del self._data[i]

    def __iter__(self) -> typing.Iterator[K]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    # end functions for ABC
    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


InstanceT = typing.TypeVar("InstanceT")
AttributeT = typing.TypeVar("AttributeT")


class IndexedContainer(typing_extensions.Protocol[T_contra]):
    """Container wth an index that can be updated."""

    def _index_discard(self, instance: T_contra) -> None:
        ...  # pragma: no cover

    def _index_add(self, instance: T_contra) -> None:
        ...  # pragma: no cover


class ParentGetter(typing_extensions.Protocol[T_contra]):
    """Interface for getting an _IndexedContainer for an instance."""

    def __call__(
        self, instance: T_contra
    ) -> typing.Optional[IndexedContainer[T_contra]]:
        ...  # pragma: no cover


class _IndexedAttribute(typing.Generic[AttributeT]):
    """
    The _IndexedAttribute descriptor notifies a parent when the attribute is
    modified. The outer class is generic in the attribute type and provides a
    __call__ method to deduce the remaining type parameters and construct the
    descriptor itself.

    Example usage:

    class Foo:
        my_int = _IndexedAttribute[int]()(lambda foo: foo.parent)
    """

    class Descriptor(typing.Generic[InstanceT]):
        """
        A descriptor that will notify a parent when the value is set and can be
        otherwise used like a normal attribute.
        """

        def __init__(self, parent_getter: ParentGetter[InstanceT]):
            self.parent_getter = parent_getter

        def __get__(
            self,
            instance: InstanceT,
            owner: typing.Type[InstanceT] = None,
        ) -> AttributeT:
            return getattr(instance, self.attribute_name)

        def __set__(self, instance: InstanceT, value: AttributeT) -> None:
            parent = self.parent_getter(instance)
            if parent:
                parent._index_discard(instance)
            setattr(instance, self.attribute_name, value)
            parent = self.parent_getter(instance)
            if parent:
                parent._index_add(instance)

        def __delete__(self, instance: InstanceT) -> None:
            raise AttributeError("can't delete attribute %s" % (self.name))

        def __set_name__(self, owner: InstanceT, name: str) -> None:
            self.name = name
            self.attribute_name = "_" + name

    def __call__(self, parent_getter: ParentGetter[InstanceT]) -> AttributeT:
        """
        Create the descriptor, but tell mypy it is the attribute type.

        The cast helps mypy recognize when the instance type satisfies a
        protocol. Mypy checks to see if the class matches the protocol instead
        of the instance. However, descriptors are treated as the descriptor
        type in the class and the attribute type in the instance. This causes
        mypy to reject the protocol, even though it works correctly at runtime.
        """
        return typing.cast(
            AttributeT, self.Descriptor[InstanceT](parent_getter)
        )


def get_desired_range(addrs: typing.Union[int, range]) -> range:
    if isinstance(addrs, int):
        return range(addrs, addrs + 1)
    else:
        return addrs


class AddrRange(typing_extensions.Protocol):
    """An object spanning a range of addresses."""

    # Protocol field types must match exactly, but properties are allowed to
    # return subtypes. This means that a class whose address or size is an int
    # will match Optional[int] properties, but not Optional[int] fields.

    @property
    def address(self) -> typing.Optional[int]:
        ...  # pragma: no cover

    @property
    def size(self) -> typing.Optional[int]:
        ...  # pragma: no cover


# Need a TypeVar bounded by the protocol so that nodes_on callers will get
# back the actual node type, not an AddrRange.
AddrRangeT = typing.TypeVar("AddrRangeT", bound=AddrRange)


def nodes_on(
    nodes: typing.Iterable[AddrRangeT],
    addrs: typing.Union[int, range],
) -> typing.Iterable[AddrRangeT]:
    desired_range = get_desired_range(addrs)
    for node in nodes:
        node_addr = node.address
        if node_addr is not None:
            node_size = node.size
            assert node_size is not None
            node_range = range(node_addr, node_addr + node_size)
            if range(
                max(desired_range.start, node_range.start),
                min(desired_range.stop, node_range.stop),
            ):
                yield node


def nodes_at(
    nodes: typing.Iterable[AddrRangeT],
    addrs: typing.Union[int, range],
) -> typing.Iterable[AddrRangeT]:
    desired_range = get_desired_range(addrs)
    for node in nodes:
        node_addr = node.address
        if node_addr is not None and node_addr in desired_range:
            yield node


def _address_interval(
    node: AddrRangeT,
) -> "typing.Optional[intervaltree.Interval[int, AddrRangeT]]":
    """
    Creates an interval tree interval based on a GTIRB node's address and
    size or returns None, if the node has no address.
    """
    node_address = node.address
    if node_address is not None:
        node_size = node.size
        assert node_size is not None
        return intervaltree.Interval(
            node_address, node_address + node_size + 1, node
        )
    else:
        return None


class OffsetRange(typing_extensions.Protocol):
    """An object spanning a range of offsets."""

    @property
    def offset(self) -> int:
        ...  # pragma: no cover

    @property
    def size(self) -> int:
        ...  # pragma: no cover


OffsetRangeT = typing.TypeVar("OffsetRangeT", bound=OffsetRange)


def _offset_interval(
    node: OffsetRangeT,
) -> "intervaltree.Interval[int, OffsetRangeT]":
    """
    Creates an interval tree interval based on a GTIRB node's offset and size.
    """
    return intervaltree.Interval(
        node.offset, node.offset + node.size + 1, node
    )


def _nodes_on_interval_tree(
    tree: "intervaltree.IntervalTree[int, AddrRangeT]",
    addrs: typing.Union[int, range],
    adjustment: int = 0,
) -> typing.Iterable[AddrRangeT]:
    """
    Implements nodes_on for an IntervalTree.
    :param tree: The IntervalTree to search.
    :param addrs: The address or addresses to locate nodes on.
    :param adjustment: An adjustment to be applied to the search range before
           consulting the interval tree.
    """

    desired_range = get_desired_range(addrs)
    for interval in tree.overlap(
        desired_range.start + adjustment, desired_range.stop + adjustment
    ):
        node = interval.data
        if node.address is None:
            continue

        # We explicitly exclude zero-sized blocks to match the existing
        # nodes_on function and prior behavior of callers before they switched
        # to using an interval tree.
        if not node.size:
            continue

        # Our interval tree ranges are closed, so we need to make sure not to
        # return items the caller didn't request.
        if node.address + node.size <= desired_range.start:
            continue

        yield node


def _nodes_at_interval_tree(
    tree: "intervaltree.IntervalTree[int, AddrRangeT]",
    addrs: typing.Union[int, range],
    adjustment: int = 0,
) -> typing.Iterable[AddrRangeT]:
    """
    Implements nodes_at for an IntervalTree.
    :param tree: The IntervalTree to search.
    :param addrs: The address or addresses to locate nodes at.
    :param adjustment: An adjustment to be applied to the search range before
           consulting the interval tree.
    """

    desired_range = get_desired_range(addrs)
    for interval in tree.overlap(
        desired_range.start + adjustment, desired_range.stop + adjustment
    ):
        # Check that it's actually in our desired range, which may have a
        # step value that excludes it. This is a constant time operation.
        if interval.data.address in desired_range:
            yield interval.data


def symbolic_expressions_at(
    nodes: typing.Iterable[_SymbolicExpressionContainer[T_cov]],
    addrs: typing.Union[int, range],
) -> typing.Iterable[T_cov]:
    return itertools.chain.from_iterable(
        node.symbolic_expressions_at(addrs) for node in nodes
    )



import io
import unittest

import gtirb.serialization
from gtirb.serialization import Variant


class TestSerialization(unittest.TestCase):
    def test_parse_type(self):
        def test_positive(type_name, oracle):
            self.assertEqual(
                gtirb.serialization.Serialization._parse_type(type_name),
                oracle,
            )

        positive_tests = [
            ("mapping", ("mapping", ())),
            ("mapping<FOO,BAR>", ("mapping", (("FOO", ()), ("BAR", ())))),
            (
                "mapping<FOO,set<BAR>>",
                ("mapping", (("FOO", ()), ("set", (("BAR", ()),)))),
            ),
            (
                "mapping<FOO,mapping<BAR,BAZ>>",
                (
                    "mapping",
                    (("FOO", ()), ("mapping", (("BAR", ()), ("BAZ", ())))),
                ),
            ),
            (
                "mapping<mapping<BAR,BAZ>,FOO>",
                (
                    "mapping",
                    (("mapping", (("BAR", ()), ("BAZ", ()))), ("FOO", ())),
                ),
            ),
        ]
        for type_name, oracle in positive_tests:
            test_positive(type_name, oracle)

        def test_negative(type_name):
            with self.assertRaises(
                gtirb.serialization.TypeNameError, msg=type_name
            ):
                gtirb.serialization.Serialization._parse_type(type_name)

        negative_tests = [
            "mapping<<>",
            "mapping<>>",
            "mapping<><>",
            "mapping<<><>>",
            "mapping<<foo><bar>>",
            "mapping<,>",
            "mapping<FOO,>",
            "mapping<,BAR>",
        ]
        for type_name in negative_tests:
            test_negative(type_name)

    def test_unknown_codec(self):
        serializer = gtirb.serialization.Serialization()
        blob = serializer.decode(b"abcd", "foobar")
        self.assertIsInstance(blob, gtirb.serialization.UnknownData)
        self.assertEqual(blob, b"abcd")
        ostream = io.BytesIO()
        serializer.encode(ostream, blob, "foobar")
        self.assertEqual(ostream.getvalue(), b"abcd")

    def test_nested_unknown_codec(self):
        serializer = gtirb.serialization.Serialization()
        ostream = io.BytesIO()
        serializer.encode(
            ostream,
            {"a": ["b", "c"], "d": ["e"]},
            "mapping<string,sequence<string>>",
        )
        raw_bytes = ostream.getvalue()

        blob = serializer.decode(raw_bytes, "mapping<string,sequence<foobar>>")
        self.assertIsInstance(blob, gtirb.serialization.UnknownData)
        ostream = io.BytesIO()
        serializer.encode(ostream, blob, "mapping<string,sequence<foobar>>")
        self.assertEqual(ostream.getvalue(), raw_bytes)

        blob = serializer.decode(raw_bytes, "mapping<foobar,sequence<string>>")
        self.assertIsInstance(blob, gtirb.serialization.UnknownData)
        ostream = io.BytesIO()
        serializer.encode(ostream, blob, "mapping<foobar,sequence<string>>")
        self.assertEqual(ostream.getvalue(), raw_bytes)

    def test_uuid_codec(self):
        ir = gtirb.IR(
            modules=[
                gtirb.Module(
                    name="M",
                    sections=[
                        gtirb.Section(
                            byte_intervals=[
                                gtirb.ByteInterval(blocks=[gtirb.CodeBlock()])
                            ]
                        )
                    ],
                )
            ]
        )
        b = next(iter(ir.code_blocks))

        # test finding block in same IR
        bstream = io.BytesIO()
        gtirb.AuxData.serializer.encode(bstream, b.uuid, "UUID")
        result = gtirb.AuxData.serializer.decode(
            bstream.getvalue(), "UUID", ir.get_by_uuid
        )
        self.assertEqual(result, b)

        # test not finding block with same UUID in other IR
        ir2 = gtirb.IR(
            modules=[
                gtirb.Module(
                    name="M",
                    sections=[
                        gtirb.Section(
                            byte_intervals=[
                                gtirb.ByteInterval(
                                    blocks=[gtirb.CodeBlock(uuid=b.uuid)]
                                )
                            ]
                        )
                    ],
                )
            ]
        )
        b2 = next(iter(ir2.code_blocks))

        bstream = io.BytesIO()
        gtirb.AuxData.serializer.encode(bstream, b.uuid, "UUID")
        result = gtirb.AuxData.serializer.decode(
            bstream.getvalue(), "UUID", ir.get_by_uuid
        )
        self.assertEqual(result, b)

        bstream = io.BytesIO()
        gtirb.AuxData.serializer.encode(bstream, b.uuid, "UUID")
        result = gtirb.AuxData.serializer.decode(
            bstream.getvalue(), "UUID", ir2.get_by_uuid
        )
        self.assertEqual(result, b2)

        # test not finding the UUID of a free block
        b3 = gtirb.CodeBlock()

        bstream = io.BytesIO()
        gtirb.AuxData.serializer.encode(bstream, b3.uuid, "UUID")
        result = gtirb.AuxData.serializer.decode(
            bstream.getvalue(), "UUID", ir.get_by_uuid
        )
        self.assertEqual(result, b3.uuid)

    def test_variant_codec(self):
        serializer = gtirb.serialization.Serialization()
        ostream = io.BytesIO()
        variant = Variant(2, {5: ["a", "b"], 15: ["cc", "ddd"]})
        serializer.encode(
            ostream,
            variant,
            "variant<string,int64_t,mapping<int64_t,sequence<string>>>",
        )
        raw_bytes = ostream.getvalue()
        var_val = serializer.decode(
            raw_bytes,
            "variant<string,int64_t,mapping<int64_t,sequence<string>>>",
        )
        self.assertEqual(var_val, variant)

        ostream = io.BytesIO()
        variant = Variant(1, 10)
        serializer.encode(
            ostream,
            variant,
            "variant<string,int64_t,string>",
        )
        raw_bytes = ostream.getvalue()
        var_val = serializer.decode(
            raw_bytes, "variant<string,int64_t,string>"
        )
        self.assertEqual(var_val, variant)

        ostream = io.BytesIO()
        variant = Variant(0, "zzzz")
        serializer.encode(
            ostream,
            variant,
            "variant<string,int64_t,string>",
        )
        raw_bytes = ostream.getvalue()
        var_val = serializer.decode(
            raw_bytes, "variant<string,int64_t,string>"
        )
        self.assertEqual(var_val, variant)

        ostream = io.BytesIO()
        mapping = {"aa": Variant(0, 5), "bbb": Variant(1, "ccccc")}
        serializer.encode(
            ostream,
            mapping,
            "mapping<string,variant<int64_t,string>>",
        )
        raw_bytes = ostream.getvalue()
        mapping_val = serializer.decode(
            raw_bytes, "mapping<string,variant<int64_t,string>>"
        )
        self.assertEqual(mapping_val, mapping)

    def _check_val(self, typename, val):
        bstream = io.BytesIO()
        gtirb.AuxData.serializer.encode(bstream, val, typename)
        result = gtirb.AuxData.serializer.decode(bstream.getvalue(), typename)
        self.assertEqual(result, val)

    def test_int_serializers(self):
        # Signed types
        def _test_range(typename, minval, maxval):
            self._check_val(typename, 127)
            self._check_val(typename, minval)
            self._check_val(typename, maxval)

        _test_range("int8_t", -(2**7), (2**7) - 1)
        _test_range("int16_t", -(2**15), (2**15) - 1)
        _test_range("int32_t", -(2**31), (2**31) - 1)
        _test_range("int64_t", -(2**63), (2**63) - 1)
        _test_range("uint8_t", 0, (2**8) - 1)
        _test_range("uint16_t", 0, (2**16) - 1)
        _test_range("uint32_t", 0, (2**32) - 1)
        _test_range("uint64_t", 0, (2**64) - 1)

    def test_float_serializers(self):
        self._check_val("float", 0.4000000059604645)
        self._check_val("double", 0.4)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock

import gtirb


class AuxDataTest(unittest.TestCase):
    def setUp(self):
        self.fake_ir = mock.MagicMock()
        self.fake_ir.get_by_uuid = None

    def test_lazy(self):
        ad1 = gtirb.AuxData("test1", "string")
        self.assertEqual(ad1.data, "test1")

        serialized = ad1._to_protobuf()
        ad2 = gtirb.AuxData._from_protobuf(serialized, self.fake_ir)
        # Peek inside: the data is not yet deserialized
        self.assertTrue(ad2._data is None)

        # Accessing the data should deserialize
        self.assertEqual(ad1.data, ad2.data)
        self.assertTrue(ad2._data is not None)

        # Just exercise repr
        self.assertEqual(
            repr(ad2), "AuxData(type_name='string', data='test1', )"
        )

    def test_lazy_never_deserialized(self):
        serialized = gtirb.AuxData("testing 123", "string")._to_protobuf()

        ad1 = gtirb.AuxData._from_protobuf(serialized, self.fake_ir)
        # Peek inside: the data is not yet deserialized
        self.assertTrue(ad1._data is None)

        serialized2 = ad1._to_protobuf()
        self.assertTrue(ad1._data is None)
        self.assertEqual(serialized, serialized2)


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import (
    TYPE_CHECKING,
    Hashable,
    Iterable,
    Iterator,
    MutableSet,
    NamedTuple,
    Optional,
    Tuple,
)
from uuid import UUID

from networkx import MultiDiGraph

from .block import CfgNode
from .proto import CFG_pb2
from .util import DeserializationError

if TYPE_CHECKING:  # pragma: no cover
    from .ir import IR


class EdgeType(Enum):
    """The type of control flow transfer indicated by a
    :class:`gtirb.Edge`.
    """

    Branch = CFG_pb2.EdgeType.Value("Type_Branch")
    """This edge is the explicit target of a jump instruction.
    May be conditional or unconditional. If conditional, there will be
    a corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`.
    """

    Call = CFG_pb2.EdgeType.Value("Type_Call")
    """This edge is the explicit target of a call instruction.
    Unless the function does not return, there will also be a
    corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`.
    """

    Fallthrough = CFG_pb2.EdgeType.Value("Type_Fallthrough")
    """This edge represents two blocks executing in sequence.
    This occurs on the non-branching paths of conditional branch
    instructions, after call instructons have returned, and when two
    blocks have no control flow between them, but another
    :class:`gtirb.Edge` targets the target block.
    If there exists a fallthrough edge from block ``A`` to block ``B``,
    then ``A`` must immediately precede ``B`` in memory.
    """

    Return = CFG_pb2.EdgeType.Value("Type_Return")
    """This edge represents a return from a function, generally via a
    return instruction. Return edges may either go to a symbolless
    :class:`gtirb.ProxyBlock`, which indicates that the set of possible
    return targets is unknown, or there may be one return edge per
    return target, which indicates that the set of possible return targets
    if fully known.
    """

    Syscall = CFG_pb2.EdgeType.Value("Type_Syscall")
    """This edge is the explicit target of a system call instruction.
    Unless the function does not return, there will also be a
    corresponding edge of type :attr:`gtirb.Edge.Type.Fallthrough`. This
    is the system call equivalent to :class:`gtirb.Edge.Type.Call`.
    """

    Sysret = CFG_pb2.EdgeType.Value("Type_Sysret")
    """This edge represents a return from a system call, generally via a
    return instruction. Return edges may either go to a symbolless
    :class:`gtirb.ProxyBlock`, which indicates that the set of possible
    return targets is unknown, or there may be one return edge per
    return target, which indicates that the set of possible return targets
    if fully known. This is the system call equivalent to
    :class:`gtirb.Edge.Type.Return`.
    """


class EdgeLabel(NamedTuple):
    """Contains a more detailed description of a :class:`gtirb.Edge`
    in the CFG.

    :ivar ~.conditional: When this edge is part of a conditional branch,
        ``conditional`` is ``True`` when the edge represents the control
        flow taken when the branch's condition is met, and ``False``
        when it represents the control flow taken when the branch's
        condition is not met. Otherwise, it is always ``False``.
    :ivar ~.direct: ``True`` if the branch or call is direct,
            and ``False`` if it is indirect. If an edge is indirect,
            then all outgoing indirect edges represent the set of
            possible locations the edge may branch to. If there
            exists an indirect outgoing edge to a :class:`gtirb.ProxyBlock`
            without any :class:`gtirb.Symbol` objects referring to it,
            then the set of all possible branch locations is unknown.
    :ivar ~.type: The type of control flow the :class:`gtirb.Edge`
        represents.
    """

    type: EdgeType
    conditional: bool = False
    direct: bool = True

    def __repr__(self) -> str:
        return (
            "Edge.Label("
            "type=Edge.Type.{type.name}, "
            "conditional={conditional!r}, "
            "direct={direct!r}, "
            ")".format(**self._asdict())
        )


class Edge(
    NamedTuple(
        "NamedTuple",
        (
            ("source", CfgNode),
            ("target", CfgNode),
            ("label", Optional[EdgeLabel]),
        ),
    )
):
    """An edge in the CFG from ``source`` to ``target``, with optional
    control-flow details in ``label``.

    :ivar ~.source: The source CFG node.
    :ivar ~.target: The target CFG node.
    :ivar ~.label: An optional label containing more control flow information.
    """

    __slots__ = ()

    def __new__(
        cls,
        source: CfgNode,
        target: CfgNode,
        label: Optional[EdgeLabel] = None,
    ) -> "Edge":
        return super().__new__(cls, source, target, label)

    Type = EdgeType
    Label = EdgeLabel


class CFG(MutableSet[Edge]):
    """A control-flow graph for an :class:`IR`. Vertices are
    :class:`CfgNode`\\s, and edges may optionally contain
    :class:`Edge.Label`\\s.

    The graph may be viewed simply as a set of :class:`Edge`\\s. For
    convenience, the :meth:`out_edges` and :meth:`in_edges` methods provide
    access to the outgoing or incoming edges of individual nodes.

    For efficency, only vertices with edges are guaranteed to be stored in this
    graph. If you want to find all vertices possible (that is, all
    :class:`CfgNode`\\s), use :meth:`IR.cfg_nodes` instead.

    Internally, the graph is stored as a NetworkX instance, which can be
    accessed using :meth:`nx`. This allows NetworkX's large library of graph
    algorithms to be used on CFGs, if desired.
    """

    def __init__(self, edges: Optional[Iterable[Edge]] = None):
        self._nxg: "MultiDiGraph[CfgNode, Hashable, Optional[EdgeLabel]]" = (
            MultiDiGraph()
        )
        if edges is not None:
            self.update(edges)

    def _edge_key(self, edge: Edge) -> Optional[Hashable]:
        if edge.source in self._nxg:
            neighbors = self._nxg[edge.source]
            if edge.target in neighbors:
                for key, e in neighbors[edge.target].items():
                    if "label" in e and e["label"] == edge.label:
                        return key
        return None

    def __contains__(self, edge: object) -> bool:
        return isinstance(edge, Edge) and self._edge_key(edge) is not None

    def __iter__(self) -> Iterator[Edge]:
        for s, t, l in self._nxg.edges(data="label"):
            yield Edge(s, t, l)

    def __len__(self) -> int:
        return len(self._nxg.edges())

    def update(self, edges: Iterable[Edge]) -> None:
        for edge in edges:
            self.add(edge)

    def add(self, edge: Edge) -> None:
        if edge not in self:
            self._nxg.add_edge(edge.source, edge.target, label=edge.label)

    def clear(self) -> None:
        self._nxg.clear()

    def discard(self, edge: Edge) -> None:
        key = self._edge_key(edge)
        if key is not None:
            self._nxg.remove_edge(edge.source, edge.target, key=key)

    def out_edges(self, node: CfgNode) -> Iterator[Edge]:
        if node in self._nxg:
            for s, t, l in self._nxg.out_edges(node, data="label"):
                yield Edge(s, t, l)

    def in_edges(self, node: CfgNode) -> Iterator[Edge]:
        if node in self._nxg:
            for s, t, l in self._nxg.in_edges(node, data="label"):
                yield Edge(s, t, l)

    @classmethod
    def _from_protobuf(
        cls, edges: Iterable[CFG_pb2.Edge], ir: Optional["IR"]
    ) -> "CFG":
        assert ir

        def make_edge(ir: "IR", edge: CFG_pb2.Edge) -> Edge:
            source_uuid = UUID(bytes=edge.source_uuid)
            source = ir.get_by_uuid(source_uuid)
            if not isinstance(source, CfgNode):
                raise DeserializationError(
                    "CFG: UUID %s is not a CfgNode" % source_uuid
                )

            target_uuid = UUID(bytes=edge.target_uuid)
            target = ir.get_by_uuid(target_uuid)
            if not isinstance(target, CfgNode):
                raise DeserializationError(
                    "CFG: UUID %s is not a CfgNode" % target_uuid
                )

            label: Optional[EdgeLabel] = None
            if edge.HasField("label"):
                label = Edge.Label(
                    Edge.Type(edge.label.type),
                    edge.label.conditional,
                    edge.label.direct,
                )

            return Edge(source, target, label)

        return CFG(make_edge(ir, edge) for edge in edges)

    def _to_protobuf(self) -> Iterable[CFG_pb2.Edge]:
        for s, t, l in self._nxg.edges(data="label"):
            proto_edge = CFG_pb2.Edge()
            proto_edge.source_uuid = s.uuid.bytes
            proto_edge.target_uuid = t.uuid.bytes
            if l:
                proto_edge.label.type = l.type.value
                proto_edge.label.conditional = l.conditional
                proto_edge.label.direct = l.direct
            yield proto_edge

    # Note: This returns a "bare" MultiDiGraph because MultiDiGraph is not
    # actually a generic type.
    def nx(self) -> MultiDiGraph:  # type: ignore[type-arg]
        return self._nxg

    def deep_eq(self, other: "CFG") -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.

        def edge_sort_key(
            edge: Edge,
        ) -> Tuple[UUID, UUID, Optional[Tuple[int, bool, bool]]]:
            label_key = -1, False, False
            if edge.label is not None:
                label_key = (
                    edge.label.type.value,
                    edge.label.conditional,
                    edge.label.direct,
                )
            return (edge.source.uuid, edge.target.uuid, label_key)

        if not isinstance(other, CFG):
            return False

        # We don't have to compare nodes for deep_eq because if an node has no
        # edges, then we do not guarantee that graphs have that node as a
        # vertex, and if it has edges, a failure of deep_eq will be detected
        # when comparing edges.

        if self._nxg.number_of_edges() != other._nxg.number_of_edges():
            return False

        self_edges = sorted(self, key=edge_sort_key)
        other_edges = sorted(other, key=edge_sort_key)

        for self_edge, other_edge in zip(self_edges, other_edges):
            if self_edge.label != other_edge.label:
                return False
            if not self_edge.source.deep_eq(other_edge.source):
                return False
            if not self_edge.target.deep_eq(other_edge.target):
                return False

        return True

    def __repr__(self) -> str:
        return "CFG(%r)" % list(self)



import unittest
import uuid

import gtirb


class DeepEqTest(unittest.TestCase):
    def test_code_block(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        b1 = gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id1)
        b2 = gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id1)
        self.assertTrue(b1.deep_eq(b2))

        b2 = gtirb.CodeBlock(size=5, decode_mode=2, offset=3, uuid=id1)
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.CodeBlock(size=1, decode_mode=5, offset=3, uuid=id1)
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.CodeBlock(size=1, decode_mode=2, offset=5, uuid=id1)
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id2)
        self.assertFalse(b1.deep_eq(b2))

    def test_data_block(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        b1 = gtirb.DataBlock(size=1, offset=3, uuid=id1)
        b2 = gtirb.DataBlock(size=1, offset=3, uuid=id1)
        self.assertTrue(b1.deep_eq(b2))

        b2 = gtirb.DataBlock(size=5, offset=3, uuid=id1)
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.DataBlock(size=1, offset=5, uuid=id1)
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.DataBlock(size=1, offset=3, uuid=id2)
        self.assertFalse(b1.deep_eq(b2))

    def test_proxy_blocks(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        b1 = gtirb.ProxyBlock(uuid=id1)
        b2 = gtirb.ProxyBlock(uuid=id1)
        self.assertTrue(b1.deep_eq(b2))

        b2 = gtirb.ProxyBlock(uuid=id2)
        self.assertFalse(b1.deep_eq(b2))

    def test_symbol(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        s1 = gtirb.Symbol(name="name", payload=None, uuid=id1)
        s2 = gtirb.Symbol(name="name", payload=None, uuid=id1)
        self.assertTrue(s1.deep_eq(s2))

        s1 = gtirb.Symbol(name="name", payload=5, uuid=id1)
        s2 = gtirb.Symbol(name="name", payload=5, uuid=id1)
        self.assertTrue(s1.deep_eq(s2))

        s1 = gtirb.Symbol(
            name="name",
            payload=gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id1),
            uuid=id1,
        )
        s2 = gtirb.Symbol(
            name="name",
            payload=gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id1),
            uuid=id1,
        )
        self.assertTrue(s1.deep_eq(s2))

        s1 = gtirb.Symbol(name="name1", payload=None, uuid=id1)
        s2 = gtirb.Symbol(name="name2", payload=None, uuid=id1)
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.Symbol(name="name", payload=None, uuid=id1)
        s2 = gtirb.Symbol(name="name", payload=5, uuid=id1)
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.Symbol(
            name="name",
            payload=gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id1),
            uuid=id1,
        )
        s2 = gtirb.Symbol(
            name="name",
            payload=gtirb.CodeBlock(size=2, decode_mode=2, offset=3, uuid=id1),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.Symbol(name="name", payload=None, uuid=id1)
        s2 = gtirb.Symbol(name="name", payload=None, uuid=id2)
        self.assertFalse(s1.deep_eq(s2))

    def test_sym_exprs(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        # SymAddrConst
        s1 = gtirb.SymAddrConst(
            offset=1,
            symbol=gtirb.Symbol(name="name", payload=None, uuid=id1),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        s2 = gtirb.SymAddrConst(
            offset=1,
            symbol=gtirb.Symbol(name="name", payload=None, uuid=id1),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        self.assertTrue(s1.deep_eq(s2))

        s1 = gtirb.SymAddrConst(
            offset=1, symbol=gtirb.Symbol(name="name", payload=None, uuid=id1)
        )
        s2 = gtirb.SymAddrConst(
            offset=2, symbol=gtirb.Symbol(name="name", payload=None, uuid=id1)
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrConst(
            offset=1, symbol=gtirb.Symbol(name="name1", payload=None, uuid=id1)
        )
        s2 = gtirb.SymAddrConst(
            offset=1, symbol=gtirb.Symbol(name="name2", payload=None, uuid=id1)
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrConst(
            offset=1,
            symbol=gtirb.Symbol(name="name", payload=None, uuid=id1),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        s2 = gtirb.SymAddrConst(
            offset=1,
            symbol=gtirb.Symbol(name="name", payload=None, uuid=id1),
        )
        self.assertFalse(s1.deep_eq(s2))

        # SymAddrAddr
        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        s2 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        self.assertTrue(s1.deep_eq(s2))

        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        s2 = gtirb.SymAddrAddr(
            offset=2,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        s2 = gtirb.SymAddrAddr(
            offset=1,
            scale=4,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        s2 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name3", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        s2 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name3", payload=None, uuid=id2),
        )
        self.assertFalse(s1.deep_eq(s2))

        s1 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
            attributes={gtirb.SymbolicExpression.Attribute.G1},
        )
        s2 = gtirb.SymAddrAddr(
            offset=1,
            scale=2,
            symbol1=gtirb.Symbol(name="name1", payload=None, uuid=id1),
            symbol2=gtirb.Symbol(name="name2", payload=None, uuid=id2),
        )
        self.assertFalse(s1.deep_eq(s2))

    def test_byte_intervals(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        id3 = uuid.uuid4()
        id4 = uuid.uuid4()
        id6 = uuid.uuid4()

        b1 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertTrue(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=None,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"1234",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=8,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=0,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=5, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    6, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                )
            },
            uuid=id1,
        )
        self.assertTrue(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                7: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id1,
        )
        self.assertFalse(b1.deep_eq(b2))

        b2 = gtirb.ByteInterval(
            address=1,
            contents=b"abcd",
            size=4,
            initialized_size=4,
            blocks=(
                gtirb.DataBlock(size=1, offset=3, uuid=id2),
                gtirb.CodeBlock(size=1, decode_mode=2, offset=3, uuid=id3),
            ),
            symbolic_expressions={
                2: gtirb.SymAddrConst(
                    3, gtirb.Symbol(name="name1", payload=4, uuid=id4)
                ),
            },
            uuid=id6,
        )
        self.assertFalse(b1.deep_eq(b2))

    def test_sections(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        id3 = uuid.uuid4()
        id4 = uuid.uuid4()

        s1 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id1,
        )
        s2 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id1,
        )
        self.assertTrue(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name2",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"12345", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name",
            byte_intervals=(gtirb.ByteInterval(contents=b"abcd", uuid=id2),),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Writable,),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(
                gtirb.Section.Flag.Readable,
                gtirb.Section.Flag.Writable,
                gtirb.Section.Flag.Loaded,
            ),
            uuid=id1,
        )
        self.assertFalse(s1.deep_eq(s2))

        s2 = gtirb.Section(
            name="name",
            byte_intervals=(
                gtirb.ByteInterval(contents=b"abcd", uuid=id2),
                gtirb.ByteInterval(contents=b"1234", uuid=id3),
            ),
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
            uuid=id4,
        )
        self.assertFalse(s1.deep_eq(s2))

    def test_cfg(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()

        e1 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=True,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertFalse(
            e1.deep_eq(
                [
                    gtirb.Edge(
                        gtirb.CodeBlock(size=1, uuid=id1),
                        gtirb.CodeBlock(size=2, uuid=id2),
                        gtirb.Edge.Label(
                            type=gtirb.Edge.Type.Branch,
                            conditional=True,
                            direct=False,
                        ),
                    )
                ]
            )
        )

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=True,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertTrue(e1.deep_eq(e2))

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=True,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertFalse(e1.deep_eq(e2))

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=3, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=True,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertFalse(e1.deep_eq(e2))

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Fallthrough,
                        conditional=True,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertFalse(e1.deep_eq(e2))

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=False,
                        direct=False,
                    ),
                )
            ]
        )
        self.assertFalse(e1.deep_eq(e2))

        e2 = gtirb.CFG(
            [
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id1),
                    gtirb.CodeBlock(size=2, uuid=id2),
                    gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=True,
                        direct=True,
                    ),
                )
            ]
        )
        self.assertFalse(e1.deep_eq(e2))

    def test_module(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        id3 = uuid.uuid4()
        id4 = uuid.uuid4()
        id5 = uuid.uuid4()
        id6 = uuid.uuid4()
        id7 = uuid.uuid4()
        id8 = uuid.uuid4()

        m1 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertTrue(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("other_value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertTrue(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="other_binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.PE,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.ARM,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="other_name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=5,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=5,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=2, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id4), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id4),),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym11", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(gtirb.Symbol(name="sym1", uuid=id5),),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect22", uuid=id8),
            ),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(gtirb.Section(name="sect2", uuid=id8),),
            uuid=id1,
        )
        self.assertFalse(m1.deep_eq(m2))

        m2 = gtirb.Module(
            aux_data={"key": gtirb.AuxData("value", "string")},
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.ELF,
            isa=gtirb.Module.ISA.X64,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            entry_point=gtirb.CodeBlock(size=1, uuid=id2),
            proxies=(gtirb.ProxyBlock(uuid=id3), gtirb.ProxyBlock(uuid=id4)),
            symbols=(
                gtirb.Symbol(name="sym1", uuid=id5),
                gtirb.Symbol(name="sym2", uuid=id6),
            ),
            sections=(
                gtirb.Section(name="sect1", uuid=id7),
                gtirb.Section(name="sect2", uuid=id8),
            ),
            uuid=id2,
        )
        self.assertFalse(m1.deep_eq(m2))

    def test_ir(self):
        id1 = uuid.uuid4()
        id2 = uuid.uuid4()
        id3 = uuid.uuid4()
        id4 = uuid.uuid4()
        id5 = uuid.uuid4()
        id6 = uuid.uuid4()
        id7 = uuid.uuid4()
        id8 = uuid.uuid4()

        ir1 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertTrue(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("other_value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertTrue(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m11", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertFalse(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(gtirb.Module(name="m1", uuid=id2),),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertFalse(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=55, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertFalse(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id1,
        )
        self.assertFalse(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=5,
            uuid=id1,
        )
        self.assertFalse(ir1.deep_eq(ir2))

        ir2 = gtirb.IR(
            modules=(
                gtirb.Module(name="m1", uuid=id2),
                gtirb.Module(name="m2", uuid=id3),
            ),
            aux_data={"key": gtirb.AuxData("value", "string")},
            cfg=(
                gtirb.Edge(
                    gtirb.CodeBlock(size=1, uuid=id4),
                    gtirb.CodeBlock(size=2, uuid=id5),
                ),
                gtirb.Edge(
                    gtirb.CodeBlock(size=3, uuid=id6),
                    gtirb.CodeBlock(size=4, uuid=id7),
                ),
            ),
            version=1,
            uuid=id8,
        )
        self.assertFalse(ir1.deep_eq(ir2))


if __name__ == "__main__":
    unittest.main()



__all__ = [
    "AuxData",
    "AuxDataContainer",
    "Block",
    "ByteBlock",
    "ByteInterval",
    "CFG",
    "CfgNode",
    "CodeBlock",
    "DataBlock",
    "Edge",
    "EdgeLabel",
    "EdgeType",
    "IR",
    "Module",
    "Node",
    "Offset",
    "ProxyBlock",
    "Section",
    "Serialization",
    "Symbol",
    "SymbolicExpression",
    "SymAddrAddr",
    "SymAddrConst",
    "Variant",
]

from .auxdata import AuxData, AuxDataContainer
from .block import Block, ByteBlock, CfgNode, CodeBlock, DataBlock, ProxyBlock
from .byteinterval import ByteInterval
from .cfg import CFG, Edge, EdgeLabel, EdgeType
from .ir import IR
from .module import Module
from .node import Node
from .offset import Offset
from .section import Section
from .serialization import Serialization, Variant
from .symbol import Symbol
from .symbolicexpression import SymAddrAddr, SymAddrConst, SymbolicExpression
from .version import API_VERSION

__version__ = API_VERSION



import unittest

import gtirb


class CFGTest(unittest.TestCase):
    def test_contains(self):
        b1, b2 = gtirb.ProxyBlock(), gtirb.ProxyBlock()
        cfg = gtirb.CFG(
            [gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch))]
        )
        self.assertFalse(
            gtirb.Edge(gtirb.ProxyBlock(), gtirb.ProxyBlock()) in cfg
        )
        self.assertFalse(
            gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Fallthrough))
            in cfg
        )
        self.assertTrue(
            gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch)) in cfg
        )

    def test_add(self):
        b1, b2 = gtirb.ProxyBlock(), gtirb.ProxyBlock()
        cfg = gtirb.CFG()
        cfg.add(gtirb.Edge(b1, b2))
        cfg.add(gtirb.Edge(b1, b2))
        cfg.add(gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch)))
        self.assertEqual(len(cfg), 2)
        self.assertTrue(gtirb.Edge(b1, b2) in cfg)
        self.assertTrue(
            gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch)) in cfg
        )

    def test_clear(self):
        cfg = gtirb.CFG(
            [
                gtirb.Edge(gtirb.ProxyBlock(), gtirb.ProxyBlock()),
                gtirb.Edge(
                    gtirb.CodeBlock(offset=0, size=1),
                    gtirb.CodeBlock(offset=1, size=2),
                ),
            ]
        )
        self.assertEqual(len(cfg), 2)

        cfg.clear()
        self.assertEqual(len(cfg), 0)

    def test_discard(self):
        b1, b2 = gtirb.ProxyBlock(), gtirb.CodeBlock(offset=0, size=1)
        cfg = gtirb.CFG(
            [
                gtirb.Edge(b1, b2),
                gtirb.Edge(
                    gtirb.ProxyBlock(), gtirb.CodeBlock(offset=1, size=2)
                ),
            ]
        )
        self.assertEqual(len(cfg), 2)

        cfg.discard(gtirb.Edge(b1, b2))
        self.assertEqual(len(cfg), 1)
        self.assertFalse(gtirb.Edge(b1, b2) in cfg)

        cfg.discard(gtirb.Edge(b1, b2))
        self.assertEqual(len(cfg), 1)
        self.assertFalse(gtirb.Edge(b1, b2) in cfg)

    def test_out_edges(self):
        b1, b2, b3 = gtirb.ProxyBlock(), gtirb.ProxyBlock(), gtirb.ProxyBlock()
        b4 = gtirb.CodeBlock(offset=0, size=1)
        cfg = gtirb.CFG(
            [
                gtirb.Edge(
                    b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Fallthrough)
                ),
                gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch)),
                gtirb.Edge(b3, b2),
            ]
        )
        self.assertEqual(sum(1 for _ in cfg.out_edges(b1)), 2)
        self.assertEqual(sum(1 for _ in cfg.out_edges(b2)), 0)
        self.assertEqual(sum(1 for _ in cfg.out_edges(b3)), 1)
        self.assertEqual(sum(1 for _ in cfg.out_edges(b4)), 0)

    def test_in_edges(self):
        b1, b2, b3 = gtirb.ProxyBlock(), gtirb.ProxyBlock(), gtirb.ProxyBlock()
        b4 = gtirb.CodeBlock(offset=0, size=1)
        cfg = gtirb.CFG(
            [
                gtirb.Edge(
                    b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Fallthrough)
                ),
                gtirb.Edge(b3, b2, gtirb.Edge.Label(gtirb.Edge.Type.Branch)),
                gtirb.Edge(b1, b3),
            ]
        )
        self.assertEqual(sum(1 for _ in cfg.in_edges(b1)), 0)
        self.assertEqual(sum(1 for _ in cfg.in_edges(b2)), 2)
        self.assertEqual(sum(1 for _ in cfg.in_edges(b3)), 1)
        self.assertEqual(sum(1 for _ in cfg.in_edges(b4)), 0)

    def test_nx(self):
        b1, b2 = gtirb.ProxyBlock(), gtirb.ProxyBlock()
        cfg = gtirb.CFG(
            [gtirb.Edge(b1, b2, gtirb.Edge.Label(gtirb.Edge.Type.Call))]
        )
        for n1, n2, lab in cfg.nx().edges(data="label"):
            self.assertEqual(n1, b1)
            self.assertEqual(n2, b2)
            self.assertEqual(lab, gtirb.Edge.Label(gtirb.Edge.Type.Call))



import unittest

import gtirb
from helpers import create_interval_etc


class BlockTest(unittest.TestCase):
    def test_contains_offset(self):
        node = gtirb.CodeBlock(offset=123, size=456, decode_mode=789)
        self.assertTrue(node.contains_offset(127))
        self.assertFalse(node.contains_offset(121))

    def test_contains_address(self):
        block = gtirb.CodeBlock(offset=123, size=456, decode_mode=789)
        byte_interval = gtirb.ByteInterval(  # noqa: F841
            address=0x0, size=579, blocks=(block,)
        )
        self.assertTrue(block.contains_address(323))  # addr: 0x143
        self.assertFalse(block.contains_address(107))  # addr: 0x6b

    def test_block_references(self):
        ir, m, s, bi = create_interval_etc(address=0x1000, size=4)
        b1 = gtirb.CodeBlock(offset=0, size=2, byte_interval=bi)
        b2 = gtirb.CodeBlock(offset=2, size=2, byte_interval=bi)

        found = set(b1.references)
        self.assertEqual(found, set())

        s1 = gtirb.Symbol(name="hello", module=m, payload=b1)
        s2 = gtirb.Symbol(name="world", module=m, payload=b2)

        found = set(b1.references)
        self.assertEqual(found, {s1})

        found = set(b2.references)
        self.assertEqual(found, {s2})

        # Change the referent to verify we update the index
        s1.referent = b2
        found = set(b1.references)
        self.assertEqual(found, set())

        found = set(b2.references)
        self.assertEqual(found, {s1, s2})

        # Discard the symbol to verify we update the index
        m.symbols.discard(s1)
        found = set(b2.references)
        self.assertEqual(found, {s2})

        # Now add it back to verify we update the index
        m.symbols.add(s1)
        found = set(b2.references)
        self.assertEqual(found, {s1, s2})

        # Then set the payload to an integer to make sure we handle that right
        s1.value = 1
        s2.value = 1
        found = set(b2.references)
        self.assertEqual(found, set())



import unittest

import gtirb


class SectionTests(unittest.TestCase):
    def test_byte_blocks(self):
        s = gtirb.Section()

        bi1 = gtirb.ByteInterval(size=4, section=s)
        bi1_code_block = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi1)
        bi1_data_block = gtirb.DataBlock(offset=1, size=1, byte_interval=bi1)

        bi2 = gtirb.ByteInterval(size=4, section=s)
        bi2_code_block = gtirb.CodeBlock(offset=0, size=1, byte_interval=bi2)
        bi2_data_block = gtirb.DataBlock(offset=1, size=1, byte_interval=bi2)

        self.assertEqual(
            set(s.byte_blocks),
            {bi1_code_block, bi1_data_block, bi2_code_block, bi2_data_block},
        )
        self.assertEqual(set(s.code_blocks), {bi1_code_block, bi2_code_block})
        self.assertEqual(set(s.data_blocks), {bi1_data_block, bi2_data_block})



import io
import struct
from re import findall
from typing import (
    Any,
    BinaryIO,
    Callable,
    Collection,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

from .node import Node
from .offset import Offset

CacheLookupFn = Optional[Callable[[UUID], Optional[Node]]]


class CodecError(Exception):
    """Base class for codec exceptions."""


class DecodeError(CodecError):
    """An exception during decoding."""


class EncodeError(CodecError):
    """An exception during encoding."""


class TypeNameError(EncodeError):
    """A type name is malformed."""

    def __init__(self, hint: str) -> None:
        super().__init__("malformed type name: '%s'" % (hint))


class UnknownCodecError(CodecError):
    """An unknown codec name is encountered.
    Caught and handled by the top-level codec methods.

    :param name: the name of the unknown codec
    """

    def __init__(self, name: str) -> None:
        self.name = name


class SubtypeTree:
    """A type hint representing a parsed serialization type name.
    A ``SubtypeTree`` is has two items: A ``str`` giving
    the name of the type and a Sequence of type parameters
    (which are also ``SubtypeTree``\\s). For example, the following are all
    valid ``SubtypeTree``\\s:

    >>> SubtypeTree('string', ())
    >>> SubtypeTree('sequence', (SubtypeTree('UUID',()),))
    >>> SubtypeTree(
        'mapping', (
            SubtypeTree('string', ()),
            SubtypeTree('set', (SubtypeTree('UUID', ()),))
        )
    )
    """

    def __init__(self, name: str, subtypes: Sequence["SubtypeTree"]) -> None:
        self.name = name
        self.subtypes = subtypes

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SubtypeTree):
            return self.name == other.name and self.subtypes == other.subtypes
        if isinstance(other, tuple):
            return (self.name, self.subtypes) == other
        return False


class Variant:
    # Because the Variant can contain arbitrary data, depending on the context
    # in which it is used, it has type Any. This requires an exception to the
    # project-wide mypy configuration that disallows Any.
    def __init__(self, index: int, val: Any):  # type: ignore[misc]
        self.index = index
        self.val = val

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Variant):
            return self.index == other.index and self.val == other.val
        return False


class Codec:
    """The base class for codecs."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
        get_by_uuid: CacheLookupFn = None,
    ) -> object:
        """Decode the specified raw data into a Python object.

        :param raw_bytes: The BytesIO object to be decoded.
        :param serialization: A Serialization instance used to invoke
            other codecs if needed.
        :param subtypes: The parsed type of this object.
        :param get_by_uuid: A function to look up nodes by UUID.
        :returns: A new Python object, as decoded from ``raw_bytes``.
        """

        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def encode(
        out: BinaryIO,
        item: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
    ) -> None:
        """Encode an item, writing the serialized object to ``out``.

        :param out: A binary stream to serialize to.
        :param item: The arbitrary Python object to encode.
        :param serialization: A Serialization instance, used to invoke
            other codecs if needed.
        :param subtypes: The parsed type of this object.
        """

        raise NotImplementedError  # pragma: no cover


class MappingCodec(Codec):
    """A Codec for mapping<K,V> entries. Implemented via ``dict``."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Mapping[object, object]:
        try:
            key_type, val_type = subtypes
        except (TypeError, ValueError):
            raise DecodeError(
                "could not unpack mapping types: %s" % str(subtypes)
            )
        mapping = dict()
        mapping_len = Uint64Codec.decode(raw_bytes)
        for _ in range(mapping_len):
            key = serialization._decode_tree(raw_bytes, key_type, get_by_uuid)
            val = serialization._decode_tree(raw_bytes, val_type, get_by_uuid)
            mapping[key] = val
        return mapping

    @staticmethod
    def encode(
        out: BinaryIO,
        mapping: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
    ) -> None:
        if not isinstance(mapping, Mapping):
            raise EncodeError("Mapping codec only supports Mappings")
        try:
            key_type, val_type = subtypes
        except (TypeError, ValueError):
            raise EncodeError(
                "could not unpack mapping types: %s" % str(subtypes)
            )
        Uint64Codec.encode(out, len(mapping))
        for key, val in mapping.items():
            serialization._encode_tree(out, key, key_type)
            serialization._encode_tree(out, val, val_type)


class OffsetCodec(Codec):
    """A Codec for :class:`gtirb.Offset` objects,
    containing a UUID and a displacement.
    """

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Offset:
        if subtypes != ():
            raise DecodeError("Offset should have no subtypes")
        element_uuid = UUIDCodec.decode(raw_bytes, get_by_uuid=get_by_uuid)
        displacement = Uint64Codec.decode(raw_bytes)

        return Offset(element_uuid, displacement)

    @staticmethod
    def encode(
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(val, Offset):
            raise EncodeError("Offset codec only supports Offsets")
        if subtypes != ():
            raise EncodeError("Offset should have no subtypes")
        UUIDCodec.encode(out, val.element_id)
        Uint64Codec.encode(out, val.displacement)


class SequenceCodec(Codec):
    """A Codec for sequence<T> entries. Implemented via ``list``."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Sequence[object]:
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise DecodeError("could not unpack sequence type: %s" % str(e))
        sequence = list()
        sequence_len = Uint64Codec.decode(raw_bytes)
        for _ in range(sequence_len):
            sequence.append(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return sequence

    @staticmethod
    def encode(
        out: BinaryIO,
        sequence: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
    ) -> None:
        if not isinstance(sequence, Sequence):
            raise EncodeError("Sequence codec only supports Collections")
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise EncodeError("could not unpack sequence type: %s" % str(e))
        Uint64Codec.encode(out, len(sequence))
        for item in sequence:
            serialization._encode_tree(out, item, subtype)


class SetCodec(Codec):
    """A Codec for set<T> entries. Implemented via ``set``."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Set[object]:
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise DecodeError("could not unpack set type: %s" % str(e))
        decoded_set = set()
        set_len = Uint64Codec.decode(raw_bytes)
        for _ in range(set_len):
            decoded_set.add(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return decoded_set

    @staticmethod
    def encode(
        out: BinaryIO,
        items: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
    ) -> None:
        if not isinstance(items, Collection):
            raise EncodeError("Set codec only supports Collections")
        try:
            (subtype,) = subtypes
        except (TypeError, ValueError) as e:
            raise EncodeError("could not unpack set type: %s" % str(e))
        Uint64Codec.encode(out, len(items))
        for item in items:
            serialization._encode_tree(out, item, subtype)


class TupleCodec(Codec):
    """A Codec for tuple<...> entries. Implemented via ``tuple``."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Tuple[object, ...]:
        # The length of a tuple is not contained in the Protobuf
        # representation, so error checking cannot be done here.
        decoded_list = list()
        for subtype in subtypes:
            decoded_list.append(
                serialization._decode_tree(raw_bytes, subtype, get_by_uuid)
            )
        return tuple(decoded_list)

    @staticmethod
    def encode(
        out: BinaryIO,
        items: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree],
    ) -> None:
        if not isinstance(items, Collection):
            raise EncodeError("Tuple codec only supports Collections")
        if len(items) != len(subtypes):
            raise EncodeError("length of tuple does not match subtype count")
        for item, subtype in zip(items, subtypes):
            serialization._encode_tree(out, item, subtype)


class StringCodec(Codec):
    """A Codec for strings."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> str:
        if subtypes != tuple():
            raise DecodeError("string should have no subtypes")
        size = Uint64Codec.decode(raw_bytes)
        return raw_bytes.read(size).decode("utf-8")

    @staticmethod
    def encode(
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(val, str):
            raise EncodeError("String codec only supports strings")
        if subtypes != ():
            raise EncodeError("string should have no subtypes")
        Uint64Codec.encode(out, len(val))
        out.write(val.encode())


class BoolCodec(Codec):
    """A Codec for bool."""

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> bool:
        if subtypes != tuple():
            raise DecodeError("bool should have no subtypes")
        return bool(raw_bytes.read(1) != b"\x00")

    @staticmethod
    def encode(
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(val, bool):
            raise EncodeError("Bool codec only supports bool")
        if subtypes != ():
            raise EncodeError("bool should have no subtypes")
        out.write(bytes([val]))


class IntegerCodec(Codec):
    """Generic base class for integer-based Codecs"""

    typname: str
    bytesize: int
    signed: bool

    @classmethod
    def decode(
        cls,
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> int:
        if subtypes != ():
            raise DecodeError(f"{cls.typname} should have no subtypes")
        return int.from_bytes(
            raw_bytes.read(cls.bytesize), byteorder="little", signed=cls.signed
        )

    @classmethod
    def encode(
        cls,
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(val, int):
            raise EncodeError("Integer codec only supports integers")
        if subtypes != ():
            raise EncodeError(f"{cls.typname} should have no subtypes")
        out.write(
            val.to_bytes(cls.bytesize, byteorder="little", signed=cls.signed)
        )


class Uint64Codec(IntegerCodec):
    """A Codec for 64-bit unsigned integers."""

    typname = "uint64_t"
    bytesize = 8
    signed = False


class Uint32Codec(IntegerCodec):
    """A Codec for 32-bit unsigned integers."""

    typname = "uint32_t"
    bytesize = 4
    signed = False


class Uint16Codec(IntegerCodec):
    """A Codec for 16-bit unsigned integers."""

    typname = "uint16_t"
    bytesize = 2
    signed = False


class Uint8Codec(IntegerCodec):
    """A Codec for 8-bit unsigned integers."""

    typname = "uint8_t"
    bytesize = 1
    signed = False


class Int64Codec(IntegerCodec):
    """A Codec for 64-bit signed integers."""

    typname = "int64_t"
    bytesize = 8
    signed = True


class Int32Codec(IntegerCodec):
    """A Codec for 32-bit signed integers."""

    typname = "int32_t"
    bytesize = 4
    signed = True


class Int16Codec(IntegerCodec):
    """A Codec for 16-bit signed integers."""

    typname = "int16_t"
    bytesize = 2
    signed = True


class Int8Codec(IntegerCodec):
    """A Codec for 8-bit signed integers."""

    typname = "int8_t"
    bytesize = 1
    signed = True


class FloatCodec(Codec):
    """Generic base class for float-based Codecs"""

    typname: str
    bytesize: int
    struct_format: str

    @classmethod
    def decode(
        cls,
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> float:
        if subtypes != ():
            raise DecodeError(f"{cls.typname} should have no subtypes")

        return struct.unpack(cls.struct_format, raw_bytes.read(cls.bytesize))[
            0
        ]

    @classmethod
    def encode(
        cls,
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(val, float):
            raise EncodeError("Float codec only supports floats")
        if subtypes != ():
            raise EncodeError(f"{cls.typname} should have no subtypes")

        out.write(struct.pack(cls.struct_format, val))


class Float32Codec(FloatCodec):
    typname = "float"
    bytesize = 4
    struct_format = "<f"


class Float64Codec(FloatCodec):
    typname = "double"
    bytesize = 8
    struct_format = "<d"


class UUIDCodec(Codec):
    """A Codec for raw UUIDs or Nodes.

    Decoding a UUID first checks the Node cache for an object with the
    corresponding UUID, and either returns the object it hits or a new
    raw UUID.
    """

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Union[UUID, Node]:
        if subtypes != ():
            raise DecodeError("UUID should have no subtypes")
        uuid = UUID(bytes=raw_bytes.read(16))
        existing_node = None if get_by_uuid is None else get_by_uuid(uuid)
        return uuid if existing_node is None else existing_node

    @staticmethod
    def encode(
        out: BinaryIO,
        val: object,
        *,
        serialization: "Serialization" = None,
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if subtypes != ():
            raise EncodeError("UUID should have no subtypes")
        if isinstance(val, Node):
            out.write(val.uuid.bytes)
        elif isinstance(val, UUID):
            out.write(val.bytes)
        else:
            raise EncodeError("UUID codec only supports UUIDs or Nodes")


class VariantCodec(Codec):
    """A Codec for variant<Ts...> entries.

    An encoded record containg two part:
    index - position of member of variant's list
    value - encoded values of selected member
    """

    @staticmethod
    def decode(
        raw_bytes: BinaryIO,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree] = (),
        get_by_uuid: Optional[CacheLookupFn] = None,
    ) -> Variant:
        index = int.from_bytes(
            raw_bytes.read(8), byteorder="little", signed=False
        )
        val = serialization._decode_tree(
            raw_bytes, subtypes[index], get_by_uuid
        )
        return Variant(index, val)

    @staticmethod
    def encode(
        out: BinaryIO,
        variant: object,
        *,
        serialization: "Serialization",
        subtypes: Sequence[SubtypeTree] = (),
    ) -> None:
        if not isinstance(variant, Variant):
            raise EncodeError("Variant codec only supports variants")
        # variant is a named tuple containg index and value
        # writing the index
        out.write(variant.index.to_bytes(8, byteorder="little"))
        # writing the value
        serialization._encode_tree(out, variant.val, subtypes[variant.index])


class UnknownData(bytes):
    """This class is a blob of bytes representing data with an unknown type.
    Generated by :func:`gtirb.Serialization.decode` when it encounters
    the name of an unknown codec. Use only at the top level of an auxdata.
    """


class Serialization:
    """Manages codecs used to serialize and deserialize GTIRB objects.

    The :meth:`gtirb.Serialization.decode` method of
    :attr:`gtirb.AuxData.serializer` is called when GTIRB AuxData is loaded via
    :meth:`gtirb.IR.load_protobuf`, and the :meth:`gtirb.Serialization.encode`
    method of :attr:`gtirb.AuxData.serializer` is called when GTIRB AuxData is
    saved to file via :meth:`gtirb.IR.save_protobuf`. You can alter the
    encoding and decoding of AuxData values via
    :attr:`gtirb.Serialization.codecs`. To do this, create a new subclass of
    :class:`gtirb.serialization.Codec` and add it to
    :attr:`gtirb.Serialization.codecs`:

    >>> gtirb.AuxData.serializer.codecs['my_custom_type'] = MyCustomCodec

    This example registers a new type name, ``my_custom_type``, and associate
    it with a new codec, ``MyCustomCodec``.

    :ivar ~.codecs: A mapping of type names to codecs. Codecs can be added
        or overridden using this dictionary.
    """

    def __init__(self) -> None:
        """Initialize with the built-in `gtirb.serialization.Codec`
        subclasses.
        """

        self.codecs: Dict[str, Type[Codec]] = {
            "Addr": Uint64Codec,
            "bool": BoolCodec,
            "Offset": OffsetCodec,
            "int64_t": Int64Codec,
            "int32_t": Int32Codec,
            "int16_t": Int16Codec,
            "int8_t": Int8Codec,
            "float": Float32Codec,
            "double": Float64Codec,
            "mapping": MappingCodec,
            "sequence": SequenceCodec,
            "set": SetCodec,
            "string": StringCodec,
            "tuple": TupleCodec,
            "uint64_t": Uint64Codec,
            "uint32_t": Uint32Codec,
            "uint16_t": Uint16Codec,
            "uint8_t": Uint8Codec,
            "UUID": UUIDCodec,
            "variant": VariantCodec,
        }

    def _decode_tree(
        self,
        raw_bytes: BinaryIO,
        type_tree: SubtypeTree,
        get_by_uuid: CacheLookupFn,
    ) -> object:
        """Decode the data in ``raw_bytes`` given a parsed type tree.

        :param raw_bytes: The binary stream to read bytes from.
        :param type_tree: The parsed type of the object encoded by
            ``raw_bytes``.
        """

        if type_tree.name not in self.codecs:
            raise UnknownCodecError(type_tree.name)
        codec = self.codecs[type_tree.name]
        return codec.decode(
            raw_bytes,
            serialization=self,
            subtypes=type_tree.subtypes,
            get_by_uuid=get_by_uuid,
        )

    def _encode_tree(
        self, out: BinaryIO, val: object, type_tree: SubtypeTree
    ) -> None:
        """Encode the data in ``val`` given a parsed type tree.

        :param out: A binary stream to write bytes to.
        :param val: The :class:`gtirb.AuxData` to encode.
        :param type_tree: The parsed type to encode ``val`` as.
        """

        if type_tree.name not in self.codecs:
            raise UnknownCodecError(type_tree.name)
        codec = self.codecs[type_tree.name]
        return codec.encode(
            out, val, serialization=self, subtypes=type_tree.subtypes
        )

    @staticmethod
    def _parse_type(type_name: str) -> SubtypeTree:
        """Given an encoded aux_data type_name, generate its parse tree.

        >>> _parse_type('foo')
        ('foo', ())

        >>> _parse_type('foo<bar>')
        ('foo', (('bar',()),))

        >>> _parse_type('foo<bar<baz>>')
        ('foo', (('bar', (('baz', ()),)),))

        :param type_name: The type name to parse into a ``SubtypeTree``.
        """
        tokens = findall("[^<>,]+|<|>|,", type_name)

        def parse(
            tokens: Sequence[str],
            tree: Iterable[SubtypeTree],
        ) -> Tuple[Tuple[SubtypeTree, ...], Sequence[None]]:
            tree = list(tree)
            # It is an error to parse nothing
            if len(tokens) == 0:
                raise TypeNameError(type_name)
            first_token, *tail = tokens

            # The first token should be a name
            if first_token in {"<", ">", ","}:
                raise TypeNameError(type_name)

            # Base case
            if len(tail) == 0:
                tree.append(SubtypeTree(first_token, ()))
                return tuple(tree), []
            next_token, *tail = tail

            # No subtypes
            if next_token == ",":
                tree.append(SubtypeTree(first_token, ()))

            # Parse subtypes
            if next_token == "<":
                # Extract just the subtype tokens and parse them
                stack = ["<"]
                subtype_tokens = list()
                remaining_tokens = list()
                for t in tail:
                    if len(stack) == 0:
                        remaining_tokens.append(t)
                        continue
                    if t == "<":
                        stack.append(t)
                    elif t == ">":
                        stack.pop()
                    subtype_tokens.append(t)
                if len(stack) > 0 or subtype_tokens[-1] != ">":
                    raise TypeNameError(type_name)
                subtypes, remaining = parse(subtype_tokens[:-1], [])
                # Parsing should consume all subtype tokens
                if len(remaining) != 0:
                    raise TypeNameError(type_name)
                tree.append(SubtypeTree(first_token, subtypes))
                # Finish if all tokens are consumed
                if len(remaining_tokens) == 0:
                    return tuple(tree), []
                next_token, *tail = remaining_tokens

            # If the next token is a comma, parse next
            if next_token == ",":
                return parse(tail, tree)

            # None of the rules match, error
            raise TypeNameError(type_name)

        # There should only be one item at the root of the tree
        try:
            (parse_tree,) = parse(tokens, [])[0]
        except ValueError:
            raise TypeNameError(type_name)
        return parse_tree

    def decode(
        self,
        raw_bytes: Union[bytes, bytearray, memoryview, BinaryIO],
        type_name: str,
        get_by_uuid: CacheLookupFn = None,
    ) -> object:
        """Decode a :class:`gtirb.AuxData` of the specified type
        from the specified byte stream.

        :param raw_bytes: The byte stream from which to read the encoded value.
        :param type_name: The type name of the object encoded by ``raw_bytes``.
        :param get_by_uuid: A function to look up nodes by UUID.
        :returns: The object encoded by ``raw_bytes``.
        """

        parse_tree = Serialization._parse_type(type_name)
        all_bytes = None
        if isinstance(raw_bytes, (bytes, bytearray, memoryview)):
            all_bytes = raw_bytes
        else:
            all_bytes = raw_bytes.read()
        try:
            return self._decode_tree(
                io.BytesIO(all_bytes), parse_tree, get_by_uuid
            )
        except UnknownCodecError:
            # we found an unknwon codec; the entire data structure can't be
            # parsed; return a blob of bytes
            return UnknownData(all_bytes)

    def encode(self, out: BinaryIO, val: object, type_name: str) -> None:
        """Encodes the value of an AuxData value to bytes.

        :param out: A binary stream to write bytes to.
        :param val: The :class:`gtirb.AuxData` to encode.
        :param type_name: The type name of the value encapsulated
            by the :class:`gtirb.AuxData`.
        """

        if isinstance(val, UnknownData):
            # it was a blob of bytes because of a decoding problem;
            # just write the whole thing out
            out.write(val)
            return
        parse_tree = Serialization._parse_type(type_name)
        try:
            self._encode_tree(out, val, parse_tree)
        except UnknownCodecError as e:
            # rethrow UnknownCodecError, because we were supposed to catch it
            # via UnknownData. This means the user manually wrote a bad type.
            raise EncodeError("unknown codec: %s" % e.name)



import glob
import imp
import setuptools
import sys
import unittest
import os
import shutil


class UseUnitTest(unittest.TestCase):
    def runTest(self):
        self.fail(
            "Running tests with setup.py is not supported because it installs "
            "dependencies incorrectly. Please use unittest directly instead."
        )


def gtirb_test_suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(UseUnitTest())
    return test_suite


# copy over files needed for sdist: README and LICENCE
root_dir = "/builds/rewriting/gtirb"
this_dir = os.path.dirname(__file__)
shutil.copyfile(os.path.join(root_dir, "README.md"), "README")
shutil.copyfile(os.path.join(root_dir, "LICENSE.txt"), "LICENSE")

# Set the version
version = imp.load_source(
    "pkginfo.version", "gtirb/version.py"
).API_VERSION

# run setuptools
if __name__ == "__main__":
    with open("README", "r") as fh:
        long_description = fh.read().replace(
            ".gtirb.svg",
            "https://raw.githubusercontent.com/"
            "GrammaTech/gtirb/master/.gtirb.svg",
        )

    setuptools.setup(
        name="gtirb",
        version=version,
        author="GrammaTech",
        author_email="gtirb@grammatech.com",
        description="GrammaTech Intermediate Representation for Binaries",
        packages=setuptools.find_packages(),
        package_data={"gtirb": ["py.typed"]},
        test_suite="setup.gtirb_test_suite",
        install_requires=[
            "networkx",
            "protobuf<=3.20.1",
            "intervaltree",
            "sortedcontainers",
            "typing-extensions>=3.7.4.2",
        ],
        extras_require={
            "doc": ["sphinx", "sphinx-autodoc-typehints"],
            "dev": ["mypy==0.961", "mypy-protobuf==3.3.0", "types-protobuf==3.20.4"],
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires=">=3.6",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/grammatech/gtirb",
        license="MIT",
    )



from typing import TYPE_CHECKING, NamedTuple, Optional, Union
from uuid import UUID

from .node import Node
from .proto import Offset_pb2
from .util import DeserializationError

if TYPE_CHECKING:  # pragma: no cover
    # Ignore flake8 "imported but unused" errors.
    from .ir import IR  # noqa: F401


class Offset(
    NamedTuple(
        "NamedTuple",
        (("element_id", Union[UUID, Node]), ("displacement", int)),
    )
):
    """
    An Offset describes a location inside a :class:`gtirb.Node`, such as a
    :class:`gtirb.DataBlock` or :class:`gtirb.ByteInterval`.

    :ivar ~.element_id: The :class:`gtirb.Node` containing the location of
            interest.
    :ivar ~.displacement: The offset inside the Node to point to.
    """

    @classmethod
    def _from_protobuf(
        cls, offset: Offset_pb2.Offset, ir: Optional["IR"]
    ) -> "Offset":
        """Decode a Protobuf object to an offset.

        :param offset: The Protobuf object.
        """

        assert ir
        element_id = UUID(bytes=offset.element_id)
        element = ir.get_by_uuid(element_id)
        if not element:
            raise DeserializationError(
                "Offset: UUID %s does not refer to a Node" % element_id
            )
        return cls(element, offset.displacement)

    def _to_protobuf(self) -> Offset_pb2.Offset:
        """Encode this offset into a Protobuf object."""

        proto_offset = Offset_pb2.Offset()
        if isinstance(self.element_id, UUID):
            proto_offset.element_id = self.element_id.bytes
        else:
            proto_offset.element_id = self.element_id.uuid.bytes
        proto_offset.displacement = self.displacement
        return proto_offset
